from typing import overload, Iterable, Generator, SupportsIndex, Self

from typediter.classes import _MutableTypedIter, Item_T
from typediter._helpers import check_types, check_object_compatibility

class TypedList_lt( _MutableTypedIter[Item_T], list[Item_T] ):
    """ Class based on list, insuring the type safety of its items

    It can be used exactly like the built-in list but its items are type-restricted,
    if the type restriction is broken, it fails with a TypeRestrictionError

    Inherits from list, but overriding methods:
        - That could insert an incompatible item

    (i) All operations returning a new instance will return a new instance of the built-in list,
    not a typed list like in the 'complete' version (TypedList)

    Attributes
    ----------
    - **i_type**
        The type restriction for the items

    Raises
    ------
    - **typediter.exceptions.TypeRestrictionError**
        Raised by a typed-checked operation to prevent
        it from inserting incompatible items in a type-safe iterable.
        (Can be raised at instance creation
        or by every operation performing type checks)
    
    - **typediter.exceptions.InvalidTypeRestrictionError**
        Raised at instance creation if the provided 'i_type':
            - is not a type
    
    - **typediter.exceptions.IterableExpectedError**
        Raised by a type-checked operation if it expected an iterable
        but received something else
        (Can be raised at instance creation
        or by every operation performing type checks)

    - **other exceptions**
        Any exception that can be raised by a built-in 'list'
        for a given operation

    Usage Example
    -------------
    >>> # importing the class
    >>> from typediter import TypedList_lt
    >>>
    >>> # creating an instance
    >>> string_list = TypedList_lt( ('A', 'B', 'C'), i_type=str )
    >>>
    >>> # supports all list operations
    >>> string_list.append( 'D' )
    >>> new_list = string_list + ['E', 'F']
    >>>
    >>> # new instances generated by non-mutating
    >>> # operations are built-ins
    >>> type( new_list ) # -> list
    >>>
    >>> # trying to insert an incompatible item in the instance,
    >>> # will raise a TypeRestrictionError
    >>> invalid_instance = TypedList_lt( ( 1, 2, 3 ), i_type=str ) # will raise TypeRestrictionError
    >>> string_list.append( 12 ) # will raise TypeRestrictionError
    >>>
    >>> # Remark:
    >>> # Operations returning new instances are not type-restricted:
    >>> mixed_list = string_list + [ 1, 2, 3 ] # -> will not fail
    """
    __module__ = 'typediter.light' # shorten class 'print' name

    # -------------------- Representation --------------------
    
    def __repr__( self ) -> str:
        typename = self.i_type.__name__
        default_representation = list(self)
        return f"(lt)List[{typename}]:{default_representation}"
    
    def __str__( self ) -> str:
        return self.__repr__()

    # -------------------- Overriding --------------------
    # -> Everything that could introduce an item of another type

    def copy( self ): # do not type hint return (allows to have correct return type on child classes)
        """ Returns a shallow copy of the typed list """
        new_list = super().copy()
        return self.__class__( new_list, i_type=self.i_type, _skip_type_check=True )
    
    def insert( self, index:SupportsIndex, obj: Item_T, / ) -> None:
        """ (TYPE-CHECKED) Insert object before index """
        check_types( obj, i_type=self.i_type )
        super().insert(index, obj)

    def append( self, obj: Item_T, / ) -> None:
        """ (TYPE-CHECKED) Append object to the end of the list """
        check_types( obj, i_type=self.i_type )
        super().append( obj )

    def extend( self, other:Iterable[Item_T], / ) -> None:
        """ (TYPE-CHECKED) Extend list by appending elements from the iterable """
        if isinstance( other, Generator ):
            # because generators can be only used once
            other = tuple(other)
        check_object_compatibility( other, i_type=self.i_type ) 
        super().extend( other )

    @overload
    def __setitem__( self, key: SupportsIndex, value: Item_T, / ) -> None: ...
    @overload
    def __setitem__( self, key: slice, value: Iterable[Item_T], / ) -> None: ...
    def __setitem__( self, key: SupportsIndex | slice, value: Item_T | Iterable[Item_T], / ) -> None:
        """ (TYPE-CHECKED) Set self[key] to value """
        if isinstance( value, Generator ):
            # because generators can be only used once
            value = tuple(value)

        if isinstance( key, slice ):
            check_object_compatibility( value, i_type=self.i_type )
        else: # SupportsIndex
            check_types( value, i_type=self.i_type )
        
        super().__setitem__( key, value ) # type: ignore[index, assignment]

    def __iadd__( self, other: Iterable[Item_T], / ) -> Self: # type: ignore[override, misc]
        """ (TYPE-CHECKED) Implement self+=value """
        if isinstance( other, Generator ):
            # because generators can be only used once
            other = tuple(other)

        check_object_compatibility( other, i_type=self.i_type )
        return super().__iadd__( other )




""" Defines helper methods for typed iterable classes """

from typing import Iterable, Generator, Hashable, AbstractSet, Optional, Any

from typediter import classes # to avoid circular imports we import module
from typediter.exceptions import (
    TypeRestrictionError,
    IterableExpectedError,
    InvalidTypeRestrictionError
)

# Type Checking

def check_types( *items:Any, i_type:type ):
    """ Private helper checking the type of a number of items 
    
    Parameters
    ----------
    - ***items**
        all the items given that needs to be checked
    
    - **i_type**
        the expected type for the items
        (comparaison is made using isinstance)

    Raises
    ------
    - **typedier.exceptions.TypeRestrictionError**
        if an item with an incomptible type is found
    """
    for item in items:
        if not isinstance( item, i_type ):
            raise TypeRestrictionError( 
                f"Type-restricted operation was expecting items of type {i_type.__name__}, "
                "but received an incompatible item:\n"
                f"({type(item).__name__}): {item}"
            )

def check_object_compatibility( other:Any, *, i_type:type ):
    """ Private helper checking if an object is a compatible iterable
    
    (i) typed iterable instances gets a fast pass because we know their items type

    Parameters
    ----------
    - **other**
        the object we want to check
    
    - **i_type**
        the expected type for its items
        ( comparaison is made using isinstance )

    Raises
    ------
    - **typedier.exceptions.TypeRestrictionError**
        if an item with an incomptible type is found in the object
    
    - **typediter.exceptions.IterableExpectedError**
        if the object is not iterable
    """
    
    if isinstance( other, (classes._FrozenTypedIter, classes._MutableTypedIter) ):
        if other.i_type != i_type:
            raise TypeRestrictionError(
                f"Type-restricted operation was expecting iterable with {i_type.__name__} type items, "
                "but received an incompatible typed iterable:\n"
                f"{other}"
            )
        return
    
    if not isinstance( other, Iterable ):
        raise IterableExpectedError( f"Operation expected iterable, but received ({type(other).__name__}): {other}" )
    
    check_types( *other, i_type=i_type )


def check_i_type( i_type: Any, *, cls:Optional[type] = None ):
    """ Private helper checking if i_type is a valid type restriction 
    
    Parameters
    ----------
    - **i_type**
        the checked type restriction

    - **cls** (optional)
        the class we will be applying the type restriction to,
        used to ensure that set/frozenset based classes have 
        a hashable type restriction
    
    Raises
    ------
    - **typedier.exceptions.InvalidTypeRestrictionError**
        if i_type is not an instance of type,
        or, for set/frozenset based classes, if the given 
        i_type isn't hashable
    """
    
    if not isinstance( i_type, type ):
        raise InvalidTypeRestrictionError(
            "Provided type restriction (i_type) is not an instance of type, "
            f"Received ({type(i_type).__name__}): {i_type}"
        )
    
    # for set/frozenset we want to be sure that the 
    # given i_type produces hashable instances
    if isinstance( cls, type ) and issubclass( cls, AbstractSet ):
        if not issubclass( i_type, Hashable ):
            raise InvalidTypeRestrictionError(
                "Provided type restriction (i_type) doesn't produce hashable instances, "
                "set/frozenset based classes need hashable items, "
                f"Received ({type(i_type).__name__}): {i_type}"
            )


# Handling the generator case

def contains_generators( iterables:Iterable[Iterable] ) -> bool:
    """ Private helper returning True if 'iterables' contains an instance of a generator
    
    (i) because of the fact we sometimes do multiple operations on the operation values, 
    before passing it to be handled by the built-in iterable method,
    and generators can be used only once, we need to find when they are present to convert them to tuples
    before using them.
    """
    for iterable in iterables:
        if isinstance( iterable, Generator ):
            return True
    return False


def unpack_generators( iterables:Iterable[Iterable] ) -> tuple[ Iterable, ... ]:
    """ Private helper converting the generators present in iterables to a tuple while not modifying the other iterables
    
    (i) because of the fact we sometimes do multiple operations on the operation values, 
    before passing it to be handled by the built-in iterable method,
    and generators can be used only once, we need to convert them to tuples before carrying on.
    """
    return tuple(
        tuple(iterable) if isinstance(iterable, Generator) else iterable for iterable in iterables
    )


""" Defines utility functions

Utility Functions
-----------------
- **is_typediter_instance**( obj )
    returns True if argument is a typed iterable instance

- **is_typediter_subclass**( cls )
    returns True if argument is a typed iterable class

- **filter_items**( items, *, i_type )
    returns a tuple of filtered items that are instances of that type

- **get_converter_func**( typed_iterable_cls, *, i_type )
    Returns a converter function taking an iterable and returning a typed iterable instance

- **get_typesafe_tuple_converter_func**( i_type )
    Returns a converter function taking an iterable and returning a type-safe built-in tuple instance

- **get_typesafe_frozenset_converter_func**( i_type )
    Returns a converter function taking an iterable and returning a type-safe built-in frozenset instance
"""

from typing import Iterable, Callable, Generator, TypeGuard, Type, TypeVar, Any, overload

from typediter.exceptions import TypeRestrictionError, IterableExpectedError
from typediter._helpers import check_i_type, check_object_compatibility
from typediter.types import TypedIterable_T
from typediter.classes.light import (
    TypedList_lt,
    TypedTuple_lt,
    TypedSet_lt,
    TypedFrozenset_lt
)
from typediter.classes.complete import (
    TypedList,
    TypedTuple,
    TypedSet,
    TypedFrozenset
)

Item_K = TypeVar( 'Item_K' )

# remark: all complete versions inherits from the light versions
TYPED_ITERABLE_BASE_CLASSES = ( TypedList_lt, TypedTuple_lt, TypedSet_lt, TypedFrozenset_lt )

def is_typediter_instance( obj: Any | TypedIterable_T ) -> TypeGuard[ TypedIterable_T ]:
    """ Returns True if the given object is a typed iterable """
    return isinstance( obj, TYPED_ITERABLE_BASE_CLASSES )

def is_typediter_subclass( cls: Any ) -> TypeGuard[ Type[TypedIterable_T] ]:
    """ Returns True if the given class is a typed iterable class """
    if isinstance( cls, type ):
        return issubclass( cls, TYPED_ITERABLE_BASE_CLASSES )
    return False


def filter_items( items:Iterable, *, i_type:Type[Item_K] ) -> tuple[ Item_K, ... ]:
    """ Returns a tuple built from the given items, containing exclusively items that are instances of the given type """
    return tuple( item for item in items if isinstance( item, i_type ) )


@overload
def get_converter_func( typed_iterable_cls:Type[TypedList], *, i_type:Type[Item_K] 
) -> Callable[ [Iterable[Item_K]], TypedList[Item_K]]: ...
@overload
def get_converter_func( typed_iterable_cls:Type[TypedList_lt], *, i_type:Type[Item_K] 
) -> Callable[ [Iterable[Item_K]], TypedList_lt[Item_K]]: ...
@overload
def get_converter_func( typed_iterable_cls:Type[TypedTuple], *, i_type:Type[Item_K] 
) -> Callable[ [Iterable[Item_K]], TypedTuple[Item_K]]: ...
@overload
def get_converter_func( typed_iterable_cls:Type[TypedTuple_lt], *, i_type:Type[Item_K] 
) -> Callable[ [Iterable[Item_K]], TypedTuple_lt[Item_K]]: ...
@overload
def get_converter_func( typed_iterable_cls:Type[TypedSet], *, i_type:Type[Item_K] 
) -> Callable[ [Iterable[Item_K]], TypedSet[Item_K]]: ...
@overload
def get_converter_func( typed_iterable_cls:Type[TypedSet_lt], *, i_type:Type[Item_K] 
) -> Callable[ [Iterable[Item_K]], TypedSet_lt[Item_K]]: ...
@overload
def get_converter_func( typed_iterable_cls:Type[TypedFrozenset], *, i_type:Type[Item_K] 
) -> Callable[ [Iterable[Item_K]], TypedFrozenset[Item_K]]: ...
@overload
def get_converter_func( typed_iterable_cls:Type[TypedFrozenset_lt], *, i_type:Type[Item_K] 
) -> Callable[ [Iterable[Item_K]], TypedFrozenset_lt[Item_K]]: ...
def get_converter_func( typed_iterable_cls:Type[TypedIterable_T], *, i_type:Type[Item_K] 
) -> Callable[ [Iterable[Item_K]], TypedIterable_T[Item_K]]:
    """ Returns a converter function taking an iterable and returning a typed iterable instance

    Parameters
    ----------
    - **typed_iterable_cls**
        typed iterable class we want the converter function to return an instance of

    - **i_type**
        The type restriction that will be applied to the 
        generated instance

    Returned Function Raises 
    ------------------------
    - **typediter.exceptions.TypeRestrictionError**
        if the given value contains incompatible items 
    
    - **typediter.exceptions.IterableExpectedError**
        if the given value is not an iterable

    - **typedier.exceptions.InvalidTypeRestrictionError**
        if the i_type given at generation time is invalid:
            - because it's not a type
            - or, for typed iterables based on 'set' or 'frozenset', if the given 'i_type' is not hashable
    
    - **other exceptions**
        Any exception that can be raised when creating an instance of 
        the built-in iterable on which the typed iterable class is based
        (list / tuple / set / frozenset)
    """

    def converter( items: Iterable[Item_K] ) -> TypedIterable_T[Item_K]:
        """ Function taking an iterable and returning a typed iterable instance 
        
        (i) function generated by typediter utility 'get_converter_func'

        Parameters
        ----------
        - **items**
            the iterable we want to convert

        Raises
        ------
        - **typediter.exceptions.TypeRestrictionError**
            if the given value contains incompatible items 
        
        - **typediter.exceptions.IterableExpectedError**
            if the given value is not an iterable

        - **typedier.exceptions.InvalidTypeRestrictionError**
            if the i_type given at generation time is invalid:
                - because it's not a type
                - or, for typed iterables based on 'set' or 'frozenset', if the given 'i_type' is not hashable
        
        - **other exceptions**
            Any exception that can be raised when creating an instance of 
            the built-in iterable on which the typed iterable class is based
            (list / tuple / set / frozenset)
        """
        return typed_iterable_cls( items, i_type=i_type )
        
    return converter

def get_typesafe_tuple_converter_func( i_type:Type[Item_K] 
) -> Callable[ [Iterable[Item_K]], tuple[Item_K, ...] ]:
    """ Returns a converter function taking an iterable and returning a type-safe built-in tuple instance 
    
    Parameters
    ----------
    - **i_type**
        the type restriction that will be applied to the 
        generated instance
    
    Returned Function Raises 
    ------------------------
    - **typediter.exceptions.TypeRestrictionError**
        if the given value contains incompatible items
    
    - **typediter.exceptions.IterableExpectedError**
        if the given value is not an iterable
    
    - **typedier.exceptions.InvalidTypeRestrictionError**
        if the i_type given at generation time is not an instance of type
    
    - **other exceptions**
        Any exception that can be raised when creating an instance of tuple
    """

    def converter( items: Iterable[Item_K] ) -> tuple[Item_K, ...]:
        """ Function converting an iterable to a type-safe built-in tuple instance 
        
        (i) function generated by typediter utility 'get_typesafe_tuple_converter_func'

        Parameters
        ----------
        - **items**
            the iterable we want to convert

        Raises
        ------
        - **typediter.exceptions.TypeRestrictionError**
            if the given value contains incompatible items
        
        - **typediter.exceptions.IterableExpectedError**
            if the given value is not an iterable
        
        - **typedier.exceptions.InvalidTypeRestrictionError**
            if the i_type given at generation time is not an instance of type
        
        - **other exceptions**
            Any exception that can be raised when creating an instance of tuple
        """
        if isinstance( items, Generator ):
            # because generators can be only used once
            items = tuple(items)
        try:
            check_i_type( i_type )
            check_object_compatibility( items, i_type=i_type )
        except (TypeRestrictionError, IterableExpectedError) as err:
            raise type(err)( 
                f"Converting value to type-safe built-in tuple failed with exception:\n"
                f"{err}"
            )
        return tuple( items )

    return converter

def get_typesafe_frozenset_converter_func( i_type:Type[Item_K] 
) -> Callable[ [Iterable[Item_K]], frozenset[Item_K] ]:
    """ Returns a converter function taking an iterable and returning a type-safe built-in frozenset instance 
    
    Parameters
    ----------
    - **i_type**
        the type restriction that will be applied to the 
        generated instance
    
    Returned Function Raises 
    ------------------------
    - **typediter.exceptions.TypeRestrictionError**
        if the given value contains incompatible items 
    
    - **typediter.exceptions.IterableExpectedError**
        if the given value is not an iterable

    - **typedier.exceptions.InvalidTypeRestrictionError**
        if the i_type given at generation time is not an instance of type, 
        or if the given 'i_type' is not hashable

    - **other exceptions**
        Any exception that can be raised when creating an instance of frozenset
    """

    def converter( items: Iterable[Item_K] ) -> frozenset[Item_K]:
        """ Function converting an iterable to a type-safe built-in frozenset instance 
        
        (i) function generated by typediter utility 'get_typesafe_frozenset_converter_func'

        Parameters
        ----------
        - **items**
            the iterable we want to convert

        Raises
        ------
        - **typediter.exceptions.TypeRestrictionError**
            if the given value contains incompatible items 
        
        - **typediter.exceptions.IterableExpectedError**
            if the given value is not an iterable

        - **typedier.exceptions.InvalidTypeRestrictionError**
            if the i_type given at generation time is not an instance of type, 
            or if the given 'i_type' is not hashable

        - **other exceptions**
            Any exception that can be raised when creating an instance of frozenset
        """
        if isinstance( items, Generator ):
            # because generators can be only used once
            items = tuple(items)
        try:
            check_i_type( i_type, cls=frozenset )
            check_object_compatibility( items, i_type=i_type )
        except (TypeRestrictionError, IterableExpectedError) as err:
            raise type(err)(
                f"Converting value to type-safe built-in frozenset failed with exception:\n"
                f"{err}"
            )
        return frozenset( items )

    return converter



from typing import overload, SupportsIndex

from typediter.classes import Item_T
from typediter.classes.light import TypedList_lt
from typediter._helpers import check_object_compatibility

class TypedList( TypedList_lt[Item_T] ):
    """ Class based on list, insuring the type safety of its items, and the type safety on operations returning a new list

    It can be used exactly like the built-in list but its items, and some operations, are type-restricted,
    if the type restriction is broken, it fails with a TypeRestrictionError

    Inherits from list, but overriding methods:
        - That could insert an incompatible item (overridden in parent 'light' version class)
        - That used to return a list, to make it return a TypedList
        (for example, list addition returns a new list instance, so TypedList addition
        returns a new TypedList instance)

    Attributes
    ----------
    - **i_type**
        The type restriction for the items

    Raises
    ------
    - **typediter.exceptions.TypeRestrictionError**
        Raised by a typed-checked operation to prevent
        it from inserting incompatible items in a type-safe iterable.
        (Can be raised at instance creation
        or by every operation performing type checks)
    
    - **typediter.exceptions.InvalidTypeRestrictionError**
        Raised at instance creation if the provided 'i_type':
            - is not a type
    
    - **typediter.exceptions.IterableExpectedError**
        Raised by a type-checked operation if it expected an iterable
        but received something else
        (Can be raised at instance creation
        or by every operation performing type checks)
    
    - **other exceptions**
        Any exception that can be raised by a built-in 'list'
        for a given operation

    Usage Example
    -------------
    >>> # importing the class
    >>> from typediter import TypedList
    >>>
    >>> # creating an instance
    >>> string_list = TypedList( ('A', 'B', 'C'), i_type=str )
    >>>
    >>> # supports all list operations
    >>> string_list.append( 'D' )
    >>> new_list = string_list + ['E', 'F']
    >>>
    >>> # new instances generated by non-mutating
    >>> # operations are also type-safe
    >>> type( new_list ) # -> TypedList
    >>>
    >>> # Remark: 
    >>> # if the non-mutating operation is handled by the built-in, a built-in is returned
    >>> builtin_addition_result = [ 'E', 'F' ] + string_list
    >>> type( builtin_addition_result ) # -> list
    >>>
    >>> # trying to insert an incompatible item in the instance,
    >>> # or in a typed iterable instance generated by an operation,
    >>> # will raise a TypeRestrictionError
    >>> invalid_instance = TypedList( ( 1, 2, 3 ), i_type=str ) # will raise TypeRestrictionError
    >>> string_list.append( 12 ) # will raise TypeRestrictionError
    >>> invalid_operation_result = string_list + [ 1, 2, 3 ] # will raise TypeRestrictionError
    """
    __module__ = 'typediter.complete' # shorten class 'print' name

    # -------------------- Representation --------------------
    
    def __repr__( self ) -> str:
        typename = self.i_type.__name__
        default_representation = list(self)
        return f"List[{typename}]:{default_representation}"

    # -------------------- Overriding --------------------
    # (mutating methods overridden in parent 'light' verion class) 
    # -> Operations handled by current class should return current class instance

    @overload
    def __getitem__( self, key: SupportsIndex, / ) -> Item_T: ...
    @overload
    def __getitem__( self, key: slice, / ) -> 'TypedList[Item_T]': ...
    def __getitem__( self, key: SupportsIndex | slice, / ) ->  Item_T | 'TypedList[Item_T]':
        """ (NO TYPE CHECKS NEEDED) Returns value for self[key]
        
        Returns:
        - for an index key: an item
        - for a slice key: a new typed list instance
        """
        if isinstance( key, slice):
            new_list = super().__getitem__( key )
            return TypedList( new_list, i_type=self.i_type, _skip_type_check=True )
        return super().__getitem__( key )

    def __add__( self, other: list[Item_T], / ) -> 'TypedList[Item_T]': # type: ignore[override]
        """ (TYPE-CHECKED) Return self+value as a new typed list """
        check_object_compatibility( other, i_type=self.i_type )
        new_list = super().__add__( other )
        return TypedList( new_list, i_type=self.i_type, _skip_type_check=True )

    def __mul__( self, value: SupportsIndex, / ) -> 'TypedList[Item_T]':
        """ (NO TYPE CHECKS NEEDED) Return self*value as a new typed list """
        new_list = super().__mul__( value )
        return TypedList( new_list, i_type=self.i_type, _skip_type_check=True )

    def __rmul__( self, value: SupportsIndex, / ) -> 'TypedList[Item_T]':
        """ (NO TYPE CHECKS NEEDED) Return value*self as a new typed list """
        new_list = super().__rmul__( value )
        return TypedList( new_list, i_type=self.i_type, _skip_type_check=True )



""" Defines utility types, for type hinting

Utility Types
-------------
- **TypedIterable_T**
    a typed iterable instance, light or complete, based on list, tuple, set, or frozenset

- **FrozenTypedIterable_T**
    a typed iterable instance, light or complete, based on tuple or frozenset

- **MutableTypedIterable_T**
    a typed iterable instance, light or complete, based on list or set
"""

from typing import TypeVar, Union

from typediter.classes.light import (
    TypedList_lt,
    TypedTuple_lt,
    TypedSet_lt,
    TypedFrozenset_lt
)
from typediter.classes.complete import (
    TypedList,
    TypedTuple,
    TypedSet,
    TypedFrozenset
)

T = TypeVar('T')

MutableTypedIterable_T = Union[
    TypedList_lt[T], 
    TypedList[T], 
    TypedSet_lt[T], 
    TypedSet[T]
]

FrozenTypedIterable_T = Union[
    TypedTuple_lt[T],
    TypedTuple[T],
    TypedFrozenset_lt[T],
    TypedFrozenset[T]
]

TypedIterable_T = Union[
    MutableTypedIterable_T[T],
    FrozenTypedIterable_T[T]
]




from typing import Iterable, AbstractSet, Generator, Self

from typediter.classes import _MutableTypedIter, Item_T
from typediter._helpers import (
    check_types,
    check_object_compatibility,
    contains_generators,
    unpack_generators
)

class TypedSet_lt( _MutableTypedIter[Item_T], set[Item_T] ):
    """ Class based on set, insuring the type safety of its items

    It can be used exactly like the built-in set but its items are type-restricted,
    if the type restriction is broken, it fails with a TypeRestrictionError

    Inherits from set, but overriding methods:
        - That could insert an incompatible item

    (i) All operations returning a new instance will return a new instance of the built-in set,
    not a typed set like in the 'complete' version (TypedSet)

    Attributes
    ----------
    - **i_type**
        The type restriction for the items

    Raises
    ------
    - **typediter.exceptions.TypeRestrictionError**
        Raised by a typed-checked operation to prevent
        it from inserting incompatible items in a type-safe iterable.
        (Can be raised at instance creation
        or by every operation performing type checks)
    
    - **typediter.exceptions.InvalidTypeRestrictionError**
        Raised at instance creation if the provided 'i_type':
            - is not a type
            - or is not a hashable type.
    
    - **typediter.exceptions.IterableExpectedError**
        Raised by a type-checked operation if it expected an iterable
        but received something else
        (Can be raised at instance creation
        or by every operation performing type checks)
    
    - **other exceptions**
        Any exception that can be raised by a built-in 'set'
        for a given operation

    Usage Example
    -------------
    >>> # importing the class
    >>> from typediter import TypedSet_lt
    >>>
    >>> # creating an instance
    >>> string_set = TypedSet_lt( ('A', 'B', 'C'), i_type=str )
    >>>
    >>> # supports all set operations
    >>> string_set.add( 'D' )
    >>> new_set = string_set & { 'E', 'F' }
    >>>
    >>> # new instances generated by non-mutating
    >>> # operations are built-ins
    >>> type( new_set ) # -> set
    >>>
    >>> # trying to insert an incompatible item in the instance,
    >>> # will raise a TypeRestrictionError
    >>> invalid_instance = TypedSet_lt( ( 1, 2, 3 ), i_type=str ) # will raise TypeRestrictionError
    >>> string_set.add( 12 ) # will raise TypeRestrictionError
    >>>
    >>> # Remark:
    >>> # Operations returning new instances are not type-restricted:
    >>> mixed_set = string_set & { 1, 2, 3 } # -> will not fail
    """
    __module__ = 'typediter.light' # shorten class 'print' name

    # -------------------- Representation --------------------
    
    def __repr__( self ) -> str:
        typename = self.i_type.__name__
        default_representation = set(self)
        return f"(lt)Set[{typename}]:{default_representation}"
    
    def __str__( self ) -> str:
        return self.__repr__()

    # -------------------- Overriding --------------------
    # -> Everything that could introduce an item of another type

    def copy( self ): # do not type hint return (allows to have correct return type on child classes)
        """ Returns a shallow copy of the typed set """
        new_set = super().copy()
        return self.__class__( new_set, i_type=self.i_type, _skip_type_check=True )
    
    def add( self, element:Item_T, / ) -> None:
        """ (TYPE-CHECKED) Add an element to a set

        This has no effect if the element is already present
        """
        check_types( element, i_type=self.i_type )
        super().add( element )
    
    def symmetric_difference_update( self, value: Iterable[Item_T], / ) -> None:
        """ (TYPE-CHECKED) Update a set with the symmetric difference of itself and another """
        if isinstance( value, Generator ):
            # because generators can be only used once
            value = tuple(value)
        
        check_object_compatibility( value, i_type=self.i_type )
        super().symmetric_difference_update( value )

    def update( self, *iterables: Iterable[Item_T] ) -> None:
        """ (TYPE-CHECKED) Update a set with the union of itself and others """
        if contains_generators(iterables):
            # because generators can be only used once
            iterables = unpack_generators( iterables )
        
        for iterable_obj in iterables:
            check_object_compatibility( iterable_obj, i_type=self.i_type )
        super().update( *iterables )
    
    def __ior__( self, value: AbstractSet[Item_T], / ) -> Self: # type: ignore[override, misc]
        """ (TYPE-CHECKED) Return self|=value """
        check_object_compatibility( value, i_type=self.i_type )
        return super().__ior__( value )

    def __ixor__( self, value: AbstractSet[Item_T], / ) -> Self: # type: ignore[override, misc]
        """ (TYPE-CHECKED) Return self^=value """
        check_object_compatibility( value, i_type=self.i_type )
        return super().__ixor__( value )




""" Module defining the "complete" version of typed iterable classes

(i) All typed iterables are extending from a built-in iterable:
list, tuple, set or frozenset.

Specificities
-------------
The "complete" version ensures the type safety of the current instance, 
and type safety of new instances generated from it.

- Mutating methods that could insert incompatible items are overridden,
they perform type checks, and will fail instead of inserting an incompatible item.

- Non-Mutating methods that used to return an instance of the same class
are overridden to return an instance of the current typed iterable,
for example, list addition returns a new list, so TypedList addition
returns a new TypedList.
So those operations will perform type checks and fail instead of inserting an 
incompatible item in the result.

The aim is too be coherent with the built-in iterables the classes are based on.
The downside is that it makes those operations heavier and less permissive.

Exported Classes
----------------
- **TypedList**
- **TypedTuple**
- **TypedSet**
- **TypedFrozenset**

Usage Exemple
-------------
>>> # importing 'complete' version of typed list class
>>> from typediter import TypedList
>>>
>>> # creating an instance
>>> string_list = TypedList( ('A', 'B', 'C'), i_type=str )
>>>
>>> # supports all the operations of the base built-in iterable
>>> string_list.append( 'D' )
>>> string_list[2] = 'Z'
>>>
>>> # new instances generated by non-mutating
>>> # operations are also type-safe
>>> new_list = string_list + ['E', 'F']
>>> type( new_list ) # -> TypedList
>>>
>>> # Remark:
>>> # if the operation is handled by the built-in, a built-in is returned
>>> builtin_addition_result = [ 'E', 'F' ] + string_list
>>> type( builtin_addition_result ) # -> list
>>>
>>> # trying to insert an incompatible item in the instance,
>>> # or in a typed iterable instance generated by an operation,
>>> # will raise a TypeRestrictionError
>>> invalid_instance = TypedList( ( 1, 2, 3 ), i_type=str ) # will raise TypeRestrictionError
>>> string_list.append( 12 ) # will raise TypeRestrictionError
>>> invalid_operation_result = string_list + [ 1, 2, 3 ] # will raise TypeRestrictionError
"""
from .list import TypedList
from .tuple import TypedTuple
from .set import TypedSet
from .frozenset import TypedFrozenset

__all__ = [
    "TypedList",
    "TypedTuple",
    "TypedSet",
    "TypedFrozenset"
]


from typediter.classes import _FrozenTypedIter, Item_T

class TypedFrozenset_lt( _FrozenTypedIter[Item_T], frozenset[Item_T] ):
    """ Class based on frozenset, insuring the type safety of its items

    It can be used exactly like the built-in frozenset but its items are type-restricted,
    if the type restriction is broken, it fails with a TypeRestrictionError
    
    Inherits from frozenset, but overriding methods:
        - That could insert an incompatible item

    (i) All operations returning a new instance will return a new instance of the built-in frozenset,
    not a typed frozenset like in the 'complete' version (TypedFrozenset)

    Attributes
    ----------
    - **i_type**
        The type restriction for the items

    Raises
    ------
    - **typediter.exceptions.TypeRestrictionError**
        Raised by a typed-checked operation to prevent
        it from inserting incompatible items in a type-safe iterable.
        (At instance creation only)
    
    - **typediter.exceptions.InvalidTypeRestrictionError**
        Raised at instance creation if the provided 'i_type':
            - is not a type
            - or is not a hashable type.
    
    - **typediter.exceptions.IterableExpectedError**
        Raised by a type-checked operation if it expected an iterable
        but received something else
        (At instance creation only)
    
    - **other exceptions**
        Any exception that can be raised by a built-in 'frozenset'
        for a given operation

    Usage Example
    -------------
    >>> # importing the class
    >>> from typediter import TypedFrozenset_lt
    >>>
    >>> # creating an instance
    >>> string_frozenset = TypedFrozenset_lt( ('A', 'B', 'C'), i_type=str )
    >>>
    >>> # supports all frozenset operations
    >>> new_frozenset = string_frozenset & { 'D', 'E', 'F' }
    >>>
    >>> # new instances generated by non-mutating
    >>> # operations are built-ins
    >>> type( new_frozenset ) # -> frozenset
    >>>
    >>> # trying to insert an incompatible item in the instance,
    >>> # will raise a TypeRestrictionError
    >>> invalid_instance = TypedFrozenset_lt( ( 1, 2, 3 ), i_type=str ) # will raise TypeRestrictionError
    >>>
    >>> # Remark:
    >>> # Operations returning new instances are not type-restricted:
    >>> mixed_frozenset = string_frozenset & { 1, 2, 3 } # -> will not fail
    """
    __module__ = 'typediter.light' # shorten class 'print' name

    # -------------------- Representation --------------------

    def __repr__( self ) -> str:
        typename = self.i_type.__name__
        default_representation = set(self)
        return f"(lt)Frozenset[{typename}]:{default_representation}"
    
    def __str__( self ) -> str:
        return self.__repr__()

    # -------------------- Overriding --------------------
    # -> Everything that could introduce an item of another type

    def copy( self ): # do not type hint return (allows to have correct return type on child classes)
        """ Returns a shallow copy of the typed frozenset """
        new_set = super().copy()
        return self.__class__( new_set, i_type=self.i_type, _skip_type_check=True )


""" Module defining the 2 sub modules each defining a flavour of typed iterable classes

Light
-----
Guarantees the type safety of the instance's items.

classes only overriding mutating methods that could insert incompatible items.
All non-mutating methods are handled by the built-in iterable they are based on,
which makes them faster. (except for copy which is overridden)

Example:
The addition between a TypedList_lt (light) and a list will return 
a new instance of a non-typed built-in list.
And the operation is not type-restricted.

(more detailed explanation in the 'light' module docstring)

Complete
--------
Guarantees the type safety of the instance's items,
as well as the type safety of the new instances generated by operations.

classes overriding mutating methods that could insert incompatible items,
as well as all methods that were returning an instance of the same class,
making them return a new typed iterable instance.

Example:
The addition between a TypedList (complete) and a list will return 
a new instance of TypedList, and if the operation can introduce an incompatible item 
in the result, it will raise a 'TypeRestrictionError'.

(more detailed explanation in the 'complete' module docstring)
"""
from ._base import (
    _FrozenTypedIter, 
    _MutableTypedIter,
    Item_T,
)

from . import light, complete

__all__ = [
    "light", "complete"
]


""" Package providing type-safe versions of some of python's built-in iterables 

Exported Classes
----------------

- lists equivalents
    - **TypedList_lt** *(light version)*
    - **TypedList** *(complete version)*

- tuples equivalents
    - **TypedTuple_lt** *(light version)*
    - **TypedTuple** *(complete version)*

- sets equivalents
    - **TypedSet_lt** *(light version)*
    - **TypedSet** *(complete version)*

- frozensets equivalents
    - **TypedFrozenset_lt** *(light version)*
    - **TypedFrozneset** *(complete version)*

    
Exported Sub-Modules
--------------------
- **utils**
    utility functions

- **exceptions**
    typediter custom exceptions

- **types**
    type hinting utilities


Difference between 'light' and 'complete'
-----------------------------------------
(i) All exported iterables are available in two different versions

**Light** (classes ending in '_lt')

Guarantees the type safety of the instance's items.

Classes only overriding mutating methods that could insert incompatible items.
All non-mutating methods are handled by the built-in iterable they are based on,
which makes them faster. (except for copy which is overridden)

*Example:*

The addition between a TypedList_lt (light) and a list will return 
a new instance of a non-typed built-in list.
And the operation is not type-restricted.


**Complete**

Guarantees the type safety of the instance's items,
as well as the type safety of the new instances generated by operations.

Classes overriding mutating methods that could insert incompatible items,
as well as all methods that were returning an instance of the same class,
making them return a new typed iterable instance.

*Example:*

The addition between a TypedList (complete) and a list will return 
a new instance of TypedList, and if the operation can introduce an incompatible item 
in the result, it can raise a 'TypeRestrictionError'.
"""

from .classes.light import (
    TypedList_lt,
    TypedTuple_lt,
    TypedSet_lt,
    TypedFrozenset_lt
)

from .classes.complete import (
    TypedList,
    TypedTuple,
    TypedSet,
    TypedFrozenset
)

from . import utils, types, exceptions

__all__ = [
    "TypedList_lt", "TypedList",
    "TypedTuple_lt", "TypedTuple",
    "TypedSet_lt", "TypedSet",
    "TypedFrozenset_lt", "TypedFrozenset",
    "utils", "exceptions", "types"
]


from typediter.classes import _FrozenTypedIter, Item_T


class TypedTuple_lt( _FrozenTypedIter[Item_T], tuple[Item_T, ...] ):
    """ Class based on tuple, insuring the type safety of its items

    It can be used exactly like the built-in tuple but its items are type-restricted,
    if the type restriction is broken, it fails with a TypeRestrictionError

    Inherits from tuple, but overriding methods:
        - That could insert an incompatible item

    (i) All operations returning a new instance will return a new instance of the built-in tuple,
    not a typed tuple like in the 'complete' version (TypedTuple)

    Attributes
    ----------
    - **i_type**
        The type restriction for the items

    Raises
    ------
    - **typediter.exceptions.TypeRestrictionError**
        Raised by a typed-checked operation to prevent
        it from inserting incompatible items in a type-safe iterable.
        (At instance creation only)
    
    - **typediter.exceptions.InvalidTypeRestrictionError**
        Raised at instance creation if the provided 'i_type':
            - is not a type
    
    - **typediter.exceptions.IterableExpectedError**
        Raised by a type-checked operation if it expected an iterable
        but received something else
        (At instance creation only)
    
    - **other exceptions**
        Any exception that can be raised by a built-in 'tuple'
        for a given operation

    Usage Example
    -------------
    >>> # importing the class
    >>> from typediter import TypedTuple_lt
    >>>
    >>> # creating an instance
    >>> string_tuple = TypedTuple_lt( ('A', 'B', 'C'), i_type=str )
    >>>
    >>> # supports all tuple operations
    >>> new_tuple = string_tuple + ('D', 'E', 'F')
    >>>
    >>> # new instances generated by non-mutating
    >>> # operations are built-ins
    >>> type( new_tuple ) # -> tuple
    >>>
    >>> # trying to insert an incompatible item in the instance,
    >>> # will raise a TypeRestrictionError
    >>> invalid_instance = TypedTuple_lt( ( 1, 2, 3 ), i_type=str ) # will raise TypeRestrictionError
    >>>
    >>> # Remark:
    >>> # Operations returning new instances are not type-restricted:
    >>> mixed_tuple = string_tuple + ( 1, 2, 3 ) # -> will not fail
    """
    __module__ = 'typediter.light' # shorten class 'print' name

    # -------------------- Representation --------------------
    
    def __repr__( self ) -> str:
        typename = self.i_type.__name__
        default_representation = tuple(self)
        return f"(lt)Tuple[{typename}]:{default_representation}"
    
    def __str__( self ) -> str:
        return self.__repr__()
    
    # -------------------- Overriding --------------------
    # -> Everything that could introduce an item of another type

    # nothing is overriten because the instance is immutable


""" Module defining the "light" version of typed iterable classes

(i) All typed iterables are extending from a built-in iterable:
list, tuple, set or frozenset.

Specificities
-------------
The "light" version (class name ending in '_lt') 
only ensures the type safety of the current instance.

- Mutating methods that could insert incompatible items are overridden,
they perform type checks, and will fail instead of inserting an incompatible item.

- Non-Mutating methods (except for 'copy' methods) are all still handled by 
the built-in class the typed iterable is based on.
So for example, a TypedList_lt addition, is not type-checked, and will return a built-in list.

The aim is to provide type safety for the current instance only, 
while avoiding slowing down operations that don't direclty affect it.
This makes those operations faster, and more permissive,
but their results are downgraded to built-in iterables.

Exported Classes
----------------
- **TypedList_lt**
- **TypedTuple_lt**
- **TypedSet_lt**
- **TypedFrozenset_lt**

Usage Example:
--------------
>>> # importing 'light' version of typed list class
>>> from typediter import TypedList_lt
>>>
>>> # creating an instance
>>> string_list = TypedList_lt( ('A', 'B', 'C'), i_type=str )
>>>
>>> # supports all the operations of the base built-in iterable
>>> string_list.append( 'D' )
>>> string_list[2] = 'Z'
>>>
>>> # new instances generated by non-mutating operations are built-ins
>>> new_list = string_list + ['E', 'F']
>>> type( new_list ) # -> list
>>>
>>> # trying to insert an incompatible item in the instance,
>>> # will raise a TypeRestrictionError
>>> invalid_instance = TypedList_lt( ( 1, 2, 3 ), i_type=str ) # will raise TypeRestrictionError
>>> string_list.append( 12 ) # will raise TypeRestrictionError
>>>
>>> # Operations returning new instances are not type-restricted:
>>> mixed_list = string_list + [ 1, 2, 3 ] # -> will not fail
"""

from .list import TypedList_lt
from .tuple import TypedTuple_lt
from .set import TypedSet_lt
from .frozenset import TypedFrozenset_lt

__all__ = [ 
    'TypedList_lt', 
    'TypedTuple_lt', 
    'TypedSet_lt', 
    'TypedFrozenset_lt'
]


""" Defines custom exceptions for the typediter package

Custom Exceptions
-----------------
- **_TypedIterError**
    Private base class for all typediter custom exceptions

- **TypeRestrictionError**
    Exception raised to prevent operation from breaking type restriction

- **OperationError**
    Base exception for errors related to mishandled type-checked operations

- **InvalidTypeRestrictionError**
    OperationError: the provided 'i_type' is invalid

- **IterableExpectedError**
    OperationError: A type-checked operation expected an iterable but received something else
"""

class _TypedIterError( Exception ):
    """ Private base class for all typediter custom exceptions """


class TypeRestrictionError( _TypedIterError ):
    """ Exception raised to prevent operation from breaking type restriction """


class OperationError( _TypedIterError ):
    """ Base exception for errors related to mishandled type-checked operations. """

class InvalidTypeRestrictionError( OperationError ):
    """ OperationError: the provided 'i_type' is invalid.
    
    Either because it's not a type, 
    or because the type restriction is applied to a set/frozenset based class
    and the provided 'i_type' is not hashable.
    """

class IterableExpectedError( OperationError ):
    """ OperationError: A type-checked operation expected an iterable but received something else """



from typing import Iterable, AbstractSet, Generator

from typediter.classes import Item_T
from typediter.classes.light import TypedSet_lt
from typediter._helpers import (
    check_object_compatibility,
    contains_generators,
    unpack_generators
)

class TypedSet( TypedSet_lt[Item_T] ):
    """ Class based on set, insuring the type safety of its items, and the type safety on operations returning a new set

    It can be used exactly like the built-in set but its items, and some operations, are type-restricted,
    if the type restriction is broken, it fails with a TypeRestrictionError

    Inherits from set, but overriding methods:
        - That could insert an incompatible item (overridden in parent 'light' version class)
        - That used to return a set, to make it return a TypedSet
        (for example, set union returns a new set instance, so TypedSet union
        returns a new TypedSet instance)

    Attributes
    ----------
    - **i_type**
        The type restriction for the items

    Raises
    ------
    - **typediter.exceptions.TypeRestrictionError**
        Raised by a typed-checked operation to prevent
        it from inserting incompatible items in a type-safe iterable.
        (Can be raised at instance creation
        or by every operation performing type checks)
    
    - **typediter.exceptions.InvalidTypeRestrictionError**
        Raised at instance creation if the provided 'i_type':
            - is not a type
            - or is not a hashable type.
    
    - **typediter.exceptions.IterableExpectedError**
        Raised by a type-checked operation if it expected an iterable
        but received something else
        (Can be raised at instance creation
        or by every operation performing type checks)
    
    - **other exceptions**
        Any exception that can be raised by a built-in 'set'
        for a given operation

    Usage Example
    -------------
    >>> # importing the class
    >>> from typediter import TypedSet
    >>>
    >>> # creating an instance
    >>> string_set = TypedSet( ('A', 'B', 'C'), i_type=str )
    >>>
    >>> # supports all set operations
    >>> string_set.add( 'D' )
    >>> new_set = string_set & { 'E', 'F' }
    >>>
    >>> # new instances generated by non-mutating
    >>> # operations are also type-safe
    >>> type( new_set ) # -> TypedSet
    >>>
    >>> # Remark: 
    >>> # if the non-mutating operation is handled by the built-in, a built-in is returned
    >>> builtin_and_result = { 'E', 'F' } & string_set
    >>> type( builtin_and_result ) # -> set
    >>>
    >>> # trying to insert an incompatible item in the instance,
    >>> # or in a typed iterable instance generated by an operation,
    >>> # will raise a TypeRestrictionError
    >>> invalid_instance = TypedSet( ( 1, 2, 3 ), i_type=str ) # will raise TypeRestrictionError
    >>> string_set.add( 12 ) # will raise TypeRestrictionError
    >>> invalid_operation_result = string_set & { 1, 2, 3 } # will raise TypeRestrictionError
    """
    __module__ = 'typediter.complete' # shorten class 'print' name

    # -------------------- Representation --------------------
    
    def __repr__( self ) -> str:
        typename = self.i_type.__name__
        default_representation = set(self)
        return f"Set[{typename}]:{default_representation}"

    # -------------------- Overriding --------------------
    # (mutating methods overridden in parent 'light' verion class) 
    # -> Operations handled by current class should return current class instance

    def difference( self, *iterables: Iterable ) -> 'TypedSet[Item_T]':
        """ (NO TYPE CHECKS NEEDED) Return the difference of two or more sets as a new typed set

        (i.e. all elements that are in this set but not the others)
        """
        # does not need type check
        # every item retured are already present in self
        new_set = super().difference( *iterables )
        return TypedSet( new_set, i_type=self.i_type, _skip_type_check=True )

    def intersection( self, *iterables: Iterable ) -> 'TypedSet[Item_T]':
        """ (NO TYPE CHECKS NEEDED) Return the intersection of two or more sets as a new typed set

        (i.e. all elements that are in all sets)
        """
        # doesn't need type check
        # if it intersects it's already in self
        new_set = super().intersection( *iterables )
        return TypedSet( new_set, i_type=self.i_type, _skip_type_check=True )
    
    def symmetric_difference( self, value: Iterable[Item_T], / ) -> 'TypedSet[Item_T]':
        """ (TYPE-CHECKED) Return the symmetric difference of two sets as a new typed set

        (i.e. all elements that are in exactly one of the sets)
        """
        if isinstance( value, Generator ):
            # because generators can be only used once
            value = tuple(value)

        check_object_compatibility( value, i_type=self.i_type )
        new_set = super().symmetric_difference( value )
        return TypedSet( new_set, i_type=self.i_type, _skip_type_check=True )

    def union( self, *iterables: Iterable[Item_T] ) -> 'TypedSet[Item_T]':  # type: ignore[override]
        """ (TYPE-CHECKED) Return the union of sets as a new typed set

        (i.e. all elements that are in either set)
        """
        if contains_generators(iterables):
            # because generators can be only used once
            iterables = unpack_generators( iterables )
            
        for iterable_obj in iterables:
            check_object_compatibility( iterable_obj, i_type=self.i_type )

        new_set = super().union( *iterables )
        return TypedSet( new_set, i_type=self.i_type, _skip_type_check=True )

    def __sub__( self, value: AbstractSet, / ) -> 'TypedSet[Item_T]':
        """ (NO TYPE CHECKS NEEDED) Return self-value as a new typed set """
        # does not need type check
        # every item retured are already present in self
        new_set = super().__sub__( value )
        return TypedSet( new_set, i_type=self.i_type, _skip_type_check=True )
    
    def __and__( self, value: AbstractSet, / ) -> 'TypedSet[Item_T]':
        """ (NO TYPE CHECKS NEEDED) Return self&value as a new typed set """
        # does not need type check
        # if it intersects it's already in self
        new_set = super().__and__( value )
        return TypedSet( new_set, i_type=self.i_type, _skip_type_check=True )
    
    def __or__( self, value: AbstractSet[Item_T], / ) -> 'TypedSet[Item_T]': # type: ignore[override]
        """ (TYPE-CHECKED) Return self|value as a new typed set """
        check_object_compatibility( value, i_type=self.i_type )
        new_set = super().__or__( value )
        return TypedSet( new_set, i_type=self.i_type, _skip_type_check=True )

    def __xor__( self, value: AbstractSet[Item_T], / ) -> 'TypedSet[Item_T]': # type: ignore[override]
        """ (TYPE-CHECKED) Return self^value as a new typed set """
        check_object_compatibility( value, i_type=self.i_type )
        new_set = super().__xor__( value )
        return TypedSet( new_set, i_type=self.i_type, _skip_type_check=True )



""" Defines base classes and types for typed iterables 

(i) all classes defined here are private and not meant to be instantiated

Defined Classes
---------------
- **_FrozenTypedIter**
    Private base class for immutable typed iterables, defining the common '__new__' method

- **_MutableTypedIter**
    Private base class for mutable typed iterables, defining the common '__init__' method

    
Utility Types
-------------
- **Item_T**
    Type bound to the iterable class, describing the type of the instance's items
"""
from enum import Enum
from typing import Iterable, Generator, TypeVar, Type, Self, Generic, Literal

from typediter._helpers import check_i_type, check_object_compatibility


class _DefaultValue(Enum):
    """ Enum class with unique 'EMPTY' value used to represent the absence of initial value at instance creation
    
    (i) we use this Enum value to represent "no initial value" in the new/init dunders instead of None
    to prevent None from being a valid initial value to create a typed iterable instance.
    """
    EMPTY = 'EMPTY'

_Empty_T = Literal[ _DefaultValue.EMPTY ]

# type bound to the iterable class, 
# describing the type of the instance's items
Item_T = TypeVar( 'Item_T' ) 


class _FrozenTypedIter( Generic[Item_T] ):
    """ Private parent class for all frozen typed iterables: 'tuple' or 'frozenset' based

    **- ⚠️ It is not meant to be instanciated. -**

    Just defines the common 'new' dunder method:
    - checking if the 'i_type' is valid
    - checking if the given items are valid and don't break 
    the given type restriction 
    
    Attributes
    ----------
    - **i_type**
        The type restriction for the iterable's items
    """
    i_type: Type[Item_T]

    def __new__( 
            cls, 
            items: Iterable | _Empty_T = _DefaultValue.EMPTY, 
            *,
            i_type: Type[Item_T], 
            _skip_type_check: bool = False 
    ) -> Self:
        """ Checks the given 'i_type' and initial 'items' before creating the instance
        
        Parameters
        ----------
        - **items** (optional) 
            The initial items given to the iterable.

        - **i_type**
            The type restriction for the iterable's items 
        
        - **_skip_type_check** (PRIVATE)
            **- ⚠️ SHOULD NOT BE USED, BREAKS THE TYPE SAFETY -**
            
            internal shortcut allowing to skip the checks for 
            instance generation from trusted items.
        
        Raises
        ------
        - **typedier.exceptions.TypeRestrictionError**
            if an item in the initial 'items' is not an instance of the given 'i_type'
        
        - **typediter.exceptions.IterableExpectedError**
            if 'items' is not an iterable
        
        - **typedier.exceptions.InvalidTypeRestrictionError**
            if 'i_type' is not an instance of type, 
            or, for 'frozenset' based classes, if 'i_type' is non-hashable
        
        - **other exceptions**
            Any exception that can be raised when creating an instance of 
            the built-in iterable on which the typed iterable class is based
            (tuple / frozenset)
        """

        # if i_type is invalid, it should fail first
        check_i_type( i_type, cls=cls )

        if items is _DefaultValue.EMPTY:
            instance = super().__new__(cls)
        else:
            if not _skip_type_check:
                if isinstance( items, Generator ):
                    # because generators can be only used once
                    items = tuple(items)
                check_object_compatibility( items, i_type=i_type )
            instance = super().__new__(cls, items)  # type: ignore[call-arg]
        
        instance.i_type = i_type
        return instance


class _MutableTypedIter( Generic[Item_T] ):
    """ Private parent class for all mutable typed iterables: 'list' or 'set' based

    **- ⚠️ It is not meant to be instanciated. -**

    Just defines the common 'init' dunder method:
    - checking if the 'i_type' is valid
    - checking if the given items are valid and and don't break 
    the given type restriction 
    
    Attributes
    ----------
    - **i_type**
        the type restriction for the iterable items
    """
    i_type: Type[Item_T]

    def __init__( 
            self, 
            items: Iterable | _Empty_T = _DefaultValue.EMPTY, 
            *, 
            i_type: Type[Item_T], 
            _skip_type_check: bool = False 
    ):
        """ Checks the given 'i_type' and initial 'items' before initialising the instance

        Parameters
        ----------
        - **items** (optional)
            the initial items given to the iterable

        - **i_type**
            The type restriction for the iterable's items 
        
        - **_skip_type_check** (PRIVATE)
            **- ⚠️ SHOULD NOT BE USED, BREAKS THE TYPE SAFETY -**

            internal shortcut allowing to skip the checks for 
            instance generation from trusted items.

        Raises
        ------
        - **typedier.exceptions.TypeRestrictionError**
            if an item in the initial 'items' is not an instance of the given 'i_type'
        
        - **typediter.exceptions.IterableExpectedError**
            if 'items' is not an iterable
        
        - **typedier.exceptions.InvalidTypeRestrictionError**
            if 'i_type' is not an instance of type, 
            or, for 'set' based classes, if 'i_type' is non-hashable
        
        - **other exceptions**
            Any exception that can be raised when creating an instance of 
            the built-in iterable on which the typed iterable class is based
            (list / set)
        """

        # if i_type is invalid, it should fail first
        check_i_type( i_type, cls=type(self) )
        self.i_type = i_type

        if items is _DefaultValue.EMPTY:
            super().__init__()
        else:
            if not _skip_type_check:
                if isinstance( items, Generator ):
                    # because generators can be only used once
                    items = tuple(items)
                check_object_compatibility( items, i_type=i_type )
            super().__init__( items ) # type: ignore[call-arg]

__all__ = [ "_FrozenTypedIter", "_MutableTypedIter" ]